@@ -38,13 +38,23 @@
 #include "privacy.h"
 #include "blist.h"
 #include "gtkutils.h"
-
-// pidgin header needed for gettext
-#define GETTEXT_PACKAGE "gtk20"
-#include <glib/gi18n-lib.h>
+#include <gtkplugin.h>
+#include <gtkprefs.h>
 
 // our auto-reply functionality
 #include "auto-reply.h"
+#include "botcheck.h"
+
+#ifdef WIN32
+#include "win32dep.h"
+#endif
+
+#ifdef ENABLE_NLS
+#include <glib/gi18n-lib.h>
+#else
+#define _(String) ((const char *) (String))
+#define N_(String) ((const char *) (String))
+#endif // ENABLE NLS
 
 static const char*
 conf_msg_unknown_autoreply()
@@ -59,6 +69,34 @@ conf_block_unknown()
 	return purple_prefs_get_bool("/plugins/core/pidgin_pp/unknown_block");
 }
 
+static gboolean
+conf_block_account_using_regex()
+{
+	return purple_prefs_get_bool(
+			"/plugins/core/pidgin_pp/block_account_with_regex");
+}
+
+static const gchar*
+conf_get_account_regex()
+{
+	return purple_prefs_get_string(
+			"/plugins/core/pidgin_pp/block_account_regex");
+}
+
+static gboolean
+conf_block_message_using_regex()
+{
+	return purple_prefs_get_bool(
+			"/plugins/core/pidgin_pp/block_message_with_regex");
+}
+
+static const gchar*
+conf_get_message_regex()
+{
+	return purple_prefs_get_string(
+			"/plugins/core/pidgin_pp/block_message_regex");
+}
+
 static gboolean
 conf_block_aol_sysmsg()
 {
@@ -89,6 +127,12 @@ conf_allow_all_irc()
 	return purple_prefs_get_bool("/plugins/core/pidgin_pp/allow_all_irc");
 }
 
+static gboolean
+conf_botcheck_enabled()
+{
+	return purple_prefs_get_bool("/plugins/core/pidgin_pp/botcheck_enable");
+}
+
 static GList*
 conf_get_block_list()
 {
@@ -168,6 +212,26 @@ msg_blocked_cb(PurpleAccount* account, char *sender)
 	}
 }
 
+#if GLIB_CHECK_VERSION(2,14,0)
+static gboolean
+pp_match_sender_regex(char *sender)
+{
+	purple_debug_info("pidgin-pp", "Block '%s' using regex?\n", sender);
+	const gchar *pattern = conf_get_account_regex();
+
+	return g_regex_match_simple(pattern, sender, 0, 0);
+}
+
+static gboolean
+pp_match_msg_regex(char *message)
+{
+	purple_debug_info("pidgin-pp", "Block '%s' using regex?\n", message);
+	const gchar *pattern = conf_get_message_regex();
+
+	return g_regex_match_simple(pattern, message, 0, 0);
+}
+#endif // GLIB_CHECK_VERSION
+
 /**
  * This is our callback for the receiving-im-msg signal.
  *
@@ -175,13 +239,21 @@ msg_blocked_cb(PurpleAccount* account, char *sender)
  */
 static gboolean
 receiving_im_msg_cb(PurpleAccount* account, char **sender, char **message,
-						int *flags, void *data)
+//						int *flags, void *data)
+			PurpleConversation *conv, PurpleMessageFlags *flags)
 {
 	PurpleBuddy* buddy;
 
 	purple_debug_info("pidgin-pp", "Got message from %s, protocol=%s\n",
 			*sender, account->protocol_id);
 
+	if (conv)
+	{
+		purple_debug_info("pidgin-pp",
+			"Message from an existing converstation, accepting\n");
+		return FALSE; // accept
+	}
+
 	// accept all IRC messages if configured accordingly
 	if ((!strcmp(account->protocol_id, "prpl-irc")) && conf_allow_all_irc())
 		return FALSE;
@@ -193,6 +265,32 @@ receiving_im_msg_cb(PurpleAccount* account, char **sender, char **message,
 		return TRUE; // block
 	}
 
+#if GLIB_CHECK_VERSION(2,14,0)
+	// block using account regex
+	if (conf_block_account_using_regex() && pp_match_sender_regex(*sender))
+	{
+		purple_debug_info(
+			"pidgin-pp", "Blocking account using regex\n");
+
+		// TODO: pidgin should actually emit a signal when we block a
+		// message (but it doesn't). remember to file a bug report.
+		msg_blocked_cb(account, *sender);
+		return TRUE; // block
+	}
+
+	// block using message regex
+	if (conf_block_message_using_regex() && pp_match_msg_regex(*message))
+	{
+		purple_debug_info(
+			"pidgin-pp", "Blocking message using regex\n");
+
+		// TODO: pidgin should actually emit a signal when we block a
+		// message (but it doesn't). remember to file a bug report.
+		msg_blocked_cb(account, *sender);
+		return TRUE; // block
+	}
+#endif // GLIB_CHECK_VERSION
+
 	// block blocked buddies
 	if (contact_is_blocked(*sender))
 	{
@@ -204,6 +302,31 @@ receiving_im_msg_cb(PurpleAccount* account, char **sender, char **message,
 		return TRUE; // block
 	}
 
+	// bot check
+	if (conf_botcheck_enabled())
+	{
+		if (botcheck_passed(*sender))
+		{
+			purple_debug_info("pidgin-pp",
+					"Botcheck: User already verified\n");
+			return FALSE; // accept
+		}
+		else if (botcheck_verify(*sender, *message))
+		{
+			botcheck_ok(account, *sender);
+
+			// we can safely block the message because the sender
+			// won't notice and we don't need to see the actual
+			// message
+			return TRUE; // block
+		}
+		else
+		{
+			botcheck_ask(account, *sender);
+			return TRUE; // block
+		}
+	}
+
 	buddy = purple_find_buddy (account, *sender);
 
 	if (buddy == NULL) // No contact list entry
@@ -232,17 +355,33 @@ receiving_im_msg_cb(PurpleAccount* account, char **sender, char **message,
 		const char* alias = purple_buddy_get_alias_only(buddy);
 		purple_debug_info("pidgin-pp", "Allowed %s\n", alias);
 	}
+
 	return FALSE; // default: accept
 }
 
 static int
+//request_authorization_cb (PurpleAccount* account, char *sender, char *msg)
 request_authorization_cb (PurpleAccount* account, char *sender)
 {
 	// < 0: deny
 	// = 0: prompt user
 	// > 0: accept
 
-	purple_debug_info("pidgin-pp", "request_authorization_cb");
+/*	// this code was intentionally left here for future use
+	if (msg != NULL)
+	{
+		purple_debug_info("pidgin-pp", "message: %s\n", msg);
+		const gchar *pattern = "test pattern";
+
+		gboolean match = g_regex_match_simple(pattern, msg, 0, 0);
+		purple_debug_info("pidgin-pp", "match: %d\n", match);
+	}
+	else
+	{
+		purple_debug_info("pidgin-pp", "NO MESSAGE\n");
+	}*/
+
+	purple_debug_info("pidgin-pp", "request_authorization_cb\n");
 
 	if (purple_prefs_get_bool("/plugins/core/pidgin_pp/block_auth_all"))
 	{
@@ -251,6 +390,15 @@ request_authorization_cb (PurpleAccount* account, char *sender)
 		return -1;
 	}
 
+	if (purple_prefs_get_bool("/plugins/core/pidgin_pp/block_auth_oscar") &&
+		(g_str_equal(account->protocol_id, "prpl-aim") ||
+		g_str_equal(account->protocol_id, "prpl-icq")))
+	{
+		purple_debug(PURPLE_DEBUG_INFO, "pidgin-pp", "Blocking "
+				"OSCAR authorization request from %s\n", sender);
+		return -1;
+	}
+
 	if (!purple_prefs_get_bool("/plugins/core/pidgin_pp/block_denied"))
 	{
 		return 0; // don't interfere, just prompt user
@@ -378,6 +526,8 @@ del_button_clicked_cb(GtkWidget *widget, GtkTreeSelection *selection)
 	GtkTreeModel *model;
 	GValue value;
 
+	memset(&value, 0, sizeof(GValue));
+
 	if (gtk_tree_selection_get_selected(selection, &model, &iter))
 	{
 		gtk_tree_model_get_value(model, &iter, 0, &value);
@@ -388,6 +538,7 @@ del_button_clicked_cb(GtkWidget *widget, GtkTreeSelection *selection)
 		{
 			gtk_tree_selection_select_iter(selection, &iter);
 		}
+		g_value_unset(&value);
 	}
 }
 
@@ -470,80 +621,128 @@ actions(PurplePlugin *plugin, gpointer context)
 	return l;
 }
 
-static PurplePluginPrefFrame *
-get_plugin_pref_frame(PurplePlugin* plugin)
+static GtkWidget *
+get_plugin_config_frame(PurplePlugin *plugin)
 {
-	PurplePluginPrefFrame *frame;
-	PurplePluginPref *ppref;
+	GtkWidget *notebook;
+	GtkWidget *config_vbox;
+	GtkWidget *tab_vbox;
+	GtkSizeGroup *sg;
 
-	frame = purple_plugin_pref_frame_new();
+	config_vbox = gtk_vbox_new(FALSE, 2);
+	gtk_container_set_border_width(GTK_CONTAINER(config_vbox), 12);
+	gtk_widget_show(config_vbox);
 
-	ppref = purple_plugin_pref_new_with_label(_("Blocked messages"));
-	purple_plugin_pref_frame_add(frame, ppref);
+	notebook = gtk_notebook_new();
+	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(notebook), GTK_POS_TOP);
+	gtk_box_pack_start(GTK_BOX(config_vbox), notebook, 0, 0, 0);
+	gtk_widget_show(notebook);
 
-	ppref = purple_plugin_pref_new_with_name_and_label
-		("/plugins/core/pidgin_pp/reply",
-		_("Auto-reply on blocked messages with:"));
-	purple_plugin_pref_frame_add(frame, ppref);
+	// Notebook page 1: Auto-reply
 
-	ppref = purple_plugin_pref_new_with_name
-				("/plugins/core/pidgin_pp/message");
-	purple_plugin_pref_frame_add(frame, ppref);
+	tab_vbox = gtk_vbox_new(FALSE, 5);
+	gtk_container_set_border_width(GTK_CONTAINER(tab_vbox), 5);
+	gtk_widget_show(tab_vbox);
+	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tab_vbox,
+			gtk_label_new(_("Auto-reply")));
 
-	ppref = purple_plugin_pref_new_with_label(_("Unknown people"));
-	purple_plugin_pref_frame_add(frame, ppref);
+	pidgin_prefs_checkbox(_("Auto-reply on blocked messages with:"),
+			"/plugins/core/pidgin_pp/reply", tab_vbox);
+	pidgin_prefs_labeled_entry(tab_vbox, "    ",
+			"/plugins/core/pidgin_pp/message", 0);
+	pidgin_prefs_checkbox(
+		_("Auto-reply on blocked messages from unknown people with:"),
+		"/plugins/core/pidgin_pp/unknown_reply", tab_vbox);
+	pidgin_prefs_labeled_entry(tab_vbox, "    ",
+			"/plugins/core/pidgin_pp/unknown_message", 0);
 
-	ppref = purple_plugin_pref_new_with_name_and_label
-		("/plugins/core/pidgin_pp/unknown_block",
-		_("Block messages from people not on your contact list"));
-	purple_plugin_pref_frame_add(frame, ppref);
+	// Notebook page 2: Messages
 
-	ppref = purple_plugin_pref_new_with_name_and_label
-		("/plugins/core/pidgin_pp/unknown_reply",
-		_("Auto-reply on blocked messages with:"));
-	purple_plugin_pref_frame_add(frame, ppref);
+	tab_vbox = gtk_vbox_new(FALSE, 5);
+	gtk_container_set_border_width(GTK_CONTAINER(tab_vbox), 5);
+	gtk_widget_show(tab_vbox);
+	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tab_vbox,
+			gtk_label_new(_("Messages")));
 
-	ppref = purple_plugin_pref_new_with_name
-				("/plugins/core/pidgin_pp/unknown_message");
-	purple_plugin_pref_frame_add(frame, ppref);
+	pidgin_prefs_checkbox(
+		_("Block messages from people not on your contact list"),
+		"/plugins/core/pidgin_pp/unknown_block", tab_vbox);
 
-	ppref = purple_plugin_pref_new_with_label(_("Authorization"));
-	purple_plugin_pref_frame_add(frame, ppref);
+#if GLIB_CHECK_VERSION(2,14,0)
+	pidgin_prefs_checkbox(
+		_("Block messages that match a regular expression:"),
+		"/plugins/core/pidgin_pp/block_message_with_regex", tab_vbox);
+	pidgin_prefs_labeled_entry(tab_vbox, "    ",
+		"/plugins/core/pidgin_pp/block_message_regex", 0);
 
-	ppref = purple_plugin_pref_new_with_name_and_label
-		("/plugins/core/pidgin_pp/block_denied",
-		_("Suppress repeated authorization requests\n(requires privacy settings to block individual users)"));
-	purple_plugin_pref_frame_add(frame, ppref);
+	pidgin_prefs_checkbox(_(
+		"Block messages from senders that match a regular expression:"),
+		"/plugins/core/pidgin_pp/block_account_with_regex", tab_vbox);
+	pidgin_prefs_labeled_entry(tab_vbox, "    ",
+		"/plugins/core/pidgin_pp/block_account_regex", 0);
+#endif // GLIB_CHECK_VERSION
 
-	ppref = purple_plugin_pref_new_with_name_and_label
-		("/plugins/core/pidgin_pp/block_auth_all",
-		_("Block all authorization requests"));
-	purple_plugin_pref_frame_add(frame, ppref);
+	// Notebook page 3: Authorization
 
-	ppref = purple_plugin_pref_new_with_name_and_label
-		("/plugins/core/pidgin_pp/auth_auto_info",
-		_("Automatically show user info on authorization requests"));
-	purple_plugin_pref_frame_add(frame, ppref);
+	tab_vbox = gtk_vbox_new(FALSE, 5);
+	gtk_container_set_border_width(GTK_CONTAINER(tab_vbox), 5);
+	gtk_widget_show(tab_vbox);
+	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tab_vbox,
+			gtk_label_new(_("Authorization")));
 
-	ppref = purple_plugin_pref_new_with_label(_("Protocol specific"));
-	purple_plugin_pref_frame_add(frame, ppref);
+	pidgin_prefs_checkbox(
+		_("Suppress repeated authorization requests\n(requires privacy settings to block individual users)"),
+		"/plugins/core/pidgin_pp/block_denied", tab_vbox);
+	pidgin_prefs_checkbox(
+		_("Block all authorization requests"),
+		"/plugins/core/pidgin_pp/block_auth_all", tab_vbox);
+	pidgin_prefs_checkbox(
+		_("Block authorization requests from OSCAR (ICQ/AIM)"),
+		"/plugins/core/pidgin_pp/block_auth_oscar", tab_vbox);
+	pidgin_prefs_checkbox(
+		_("Automatically show user info on authorization requests"),
+		"/plugins/core/pidgin_pp/auth_auto_info", tab_vbox);
 
-	ppref = purple_plugin_pref_new_with_name_and_label
-		("/plugins/core/pidgin_pp/block_jabber_headlines",
-		_("Block jabber headline messages\n(eg. MSN alerts, announcements etc.)"));
-	purple_plugin_pref_frame_add(frame, ppref);
+	// Notebook page 4: Bot check
 
-	ppref = purple_plugin_pref_new_with_name_and_label
-		("/plugins/core/pidgin_pp/allow_all_irc",
-		_("Allow all messages on IRC"));
-	purple_plugin_pref_frame_add(frame, ppref);
+	tab_vbox = gtk_vbox_new(FALSE, 5);
+	gtk_container_set_border_width(GTK_CONTAINER(tab_vbox), 5);
+	gtk_widget_show(tab_vbox);
+	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tab_vbox,
+			gtk_label_new(_("Bot check")));
 
-	ppref = purple_plugin_pref_new_with_name_and_label
-		("/plugins/core/pidgin_pp/block_aol_sys",
-		_("Block AOL system messages"));
-	purple_plugin_pref_frame_add(frame, ppref);
+	sg = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
 
-	return frame;
+	pidgin_prefs_checkbox(
+		_("Verify message sender by asking a question"),
+		"/plugins/core/pidgin_pp/botcheck_enable", tab_vbox);
+	pidgin_prefs_labeled_entry(tab_vbox, _("Question:"),
+		"/plugins/core/pidgin_pp/botcheck_question", sg);
+	pidgin_prefs_labeled_entry(tab_vbox, _("Answer:"),
+		"/plugins/core/pidgin_pp/botcheck_answer", sg);
+	pidgin_prefs_labeled_entry(tab_vbox, _("OK message:"),
+		"/plugins/core/pidgin_pp/botcheck_ok", sg);
+
+	// Notebook page 5: Protocol specific
+
+	tab_vbox = gtk_vbox_new(FALSE, 5);
+	gtk_container_set_border_width(GTK_CONTAINER(tab_vbox), 5);
+	gtk_widget_show(tab_vbox);
+	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tab_vbox,
+			gtk_label_new(_("Protocol specific")));
+
+
+	pidgin_prefs_checkbox(
+		_("Block jabber headline messages (MSN alerts, announcements etc.)"),
+		"/plugins/core/pidgin_pp/block_jabber_headlines", tab_vbox);
+	pidgin_prefs_checkbox(
+		_("Allow all messages on IRC"),
+		"/plugins/core/pidgin_pp/allow_all_irc", tab_vbox);
+	pidgin_prefs_checkbox(
+		_("Block AOL system messages"),
+		"/plugins/core/pidgin_pp/block_aol_sys", tab_vbox);
+
+	return config_vbox;
 }
 
 static gboolean
@@ -564,12 +763,26 @@ plugin_load (PurplePlugin * plugin)
 	purple_prefs_add_bool("/plugins/core/pidgin_pp/block_aol_sys", FALSE);
 	purple_prefs_add_string("/plugins/core/pidgin_pp/message",
 				_("Your message could not be delivered"));
-	purple_prefs_add_string ("/plugins/core/pidgin_pp/unknown_message",
+	purple_prefs_add_string("/plugins/core/pidgin_pp/unknown_message",
 		_("I currently only accept messages from people on my contact"
 				" list - please request my authorization."));
 	purple_prefs_add_bool("/plugins/core/pidgin_pp/block_denied", FALSE);
 	purple_prefs_add_bool("/plugins/core/pidgin_pp/block_auth_all", FALSE);
+	purple_prefs_add_bool("/plugins/core/pidgin_pp/block_auth_oscar", FALSE);
+	purple_prefs_add_bool("/plugins/core/pidgin_pp/block_account_with_regex", FALSE);
+	purple_prefs_add_string("/plugins/core/pidgin_pp/block_account_regex",
+			"spam.*bot");
+	purple_prefs_add_bool("/plugins/core/pidgin_pp/block_message_with_regex", FALSE);
+	purple_prefs_add_string("/plugins/core/pidgin_pp/block_message_regex",
+			"(leather jackets?|gold watch)");
 	purple_prefs_add_string_list("/plugins/core/pidgin_pp/block", NULL);
+	purple_prefs_add_bool("/plugins/core/pidgin_pp/botcheck_enable", FALSE);
+	purple_prefs_add_string("/plugins/core/pidgin_pp/botcheck_question",
+		_("To prove that you are human, please enter the result of 8+3"));
+	purple_prefs_add_string("/plugins/core/pidgin_pp/botcheck_answer",
+		_("11"));
+	purple_prefs_add_string("/plugins/core/pidgin_pp/botcheck_ok",
+		_("Very well then, you may speak"));
 
 	purple_signal_connect(conv_handle, "receiving-im-msg",
 			plugin, PURPLE_CALLBACK (receiving_im_msg_cb), NULL);
@@ -607,7 +820,17 @@ plugin_unload(PurplePlugin * plugin)
 	return TRUE;
 }
 
-static PurplePluginUiInfo prefs_info = { get_plugin_pref_frame };
+static PidginPluginUiInfo ui_info =
+{
+	get_plugin_config_frame,
+	0, // page_num (reserved)
+
+	// padding
+	NULL,
+	NULL,
+	NULL,
+	NULL
+};
 
 static PurplePluginInfo info =
 {
@@ -615,17 +838,17 @@ static PurplePluginInfo info =
 	PURPLE_MAJOR_VERSION,
 	PURPLE_MINOR_VERSION,
 	PURPLE_PLUGIN_STANDARD,				/**< type           */
-	NULL,						/**< ui_requirement */
+	PIDGIN_PLUGIN_TYPE,				/**< ui_requirement */
 	0,						/**< flags          */
 	NULL,						/**< dependencies   */
 	PURPLE_PRIORITY_DEFAULT,			/**< priority       */
 
 	N_("core-pidgin_pp_"),				/**< id             */
-	N_("Privacy Please"),				/**< name           */
-	VERSION,					/**< version        */
-	N_("Stops IM-spam"),				/**< summary	    */
+	NULL,						/**< name           */
+	PACKAGE_VERSION,				/**< version        */
+	NULL,						/**< summary	    */
 							/**  description    */
-	N_("A simple plugin to stop unwanted messages and repeated authorization requests from spammers."),
+	NULL,
 	"Stefan Ott <stefan@ott.net>",			/**< author         */
 	"http://pidgin-privacy-please.googlecode.com/",	/**< homepage       */
 
@@ -633,15 +856,29 @@ static PurplePluginInfo info =
 	plugin_unload,					/**< unload         */
 	NULL,						/**< destroy        */
 
-	NULL,						/**< ui_info        */
+	&ui_info,					/**< ui_info        */
 	NULL,						/**< extra_info     */
-	&prefs_info,					/**< prefs_info     */
-	actions
+	NULL,                                           /**< prefs_info     */
+	actions,
+
+	// padding
+	NULL,
+	NULL,
+	NULL,
+	NULL
 };
 
 static void
 init_plugin(PurplePlugin * plugin)
 {
+#ifdef ENABLE_NLS
+	bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
+	bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
+#endif /* ENABLE_NLS */
+	info.name        = _("Privacy Please");
+	info.summary     = _("Stops IM-spam");
+	info.description = _("A simple plugin to stop unwanted messages and repeated authorization requests from spammers.");
+
 	purple_prefs_add_none("/plugins");
 	purple_prefs_add_none("/plugins/core");
 	purple_prefs_add_none("/plugins/core/pidgin_pp");
